/* --------------------------------
   Phantasie II CD32
   Game saving
   --------------------------------
   Copyright (C) 2010-2016 Franck Charlet.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:
   
       1. Redistributions of source code must retain the above copyright notice,
          this list of conditions and the following disclaimer.
       2. Redistributions in binary form must reproduce the above copyright notice,
          this list of conditions and the following disclaimer in the documentation
          and/or other materials provided with the distribution.

   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS "AS IS" AND
   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
   ARE DISCLAIMED. IN NO EVENT SHALL FRANCK CHARLET OR CONTRIBUTORS BE LIABLE
   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
   SUCH DAMAGE.
   -------------------------------- */

/* TODO: - Save more than one game (there's room for ~3 in the nvram)
*/

/* --------------------------------
   Includes
   -------------------------------- */
#include "main.h"
#include "towns.h"
#include "save.h"

/* --------------------------------
   Variables
   -------------------------------- */
LPMENU Menu_Ask_Saving;
LPMENU Menu_Saving;
LPMENU Menu_NVRAM_Warning;

SAVEDAT Save;
UBYTE *Crunched_Buffer;

UBYTE Restored_Game[2048];

BOOL Can_Save = TRUE;
BOOL Saved_Game = FALSE;

/* --------------------------------
   Wait for player to save the game or not
   -------------------------------- */
void Ask_Saving_Events(void *Self, int Item, ULONG Pad_Dats)
{
    int i;
    int Size_Crunched;

    switch(Pad_Dats)
    {
        case PAD_OK:

            Hide_Menu(Menu_Ask_Saving, Pointer);
            Destroy_Menu(Menu_Ask_Saving);

            if(Get_Selected_Menu_Column((LPMENU) Self) == 0)
            {
                /* Do save the game */
                Menu_Saving = Construct_Menu(Font,
                                             Pic_Menus,
                                             NULL,
                                             MENU_TEXT_CENTER | MENU_NO_POINTER,
                                             4, 0,
                                             1, 0,
                                             MOVE_NONE,
                                             "SAVING GAME..."
                                            );
                Show_Menu(Menu_Saving, Pointer, -1, 205, 0);

                memcpy(&Save.Party, &Party, sizeof(PARTY));
                /* Save the current stock of all towns */
                for(i = 0; i < MAX_TOWNS; i++)
                {
                    memcpy(&Save.Towns_Stock[i][0], &Towns[i].Stock, MAX_ITEMS);
                }

                /* Pack it */
                Size_Crunched = sizeof(SAVEDAT);
                Crunched_Buffer = Pack_Data((UBYTE *) &Save, &Size_Crunched);

                /* Save the packed data into the NVRAM */
                if(NVRAM_Save("P", "2", Crunched_Buffer, Size_Crunched))
                {
                    /* Notify the player if there's any error */
                    Display_Message(Font, Pic_Menus, 205, "ERROR WHILE SAVING GAME\n%d BYTES NEEDED IN NVRAM", Size_Crunched);
                }

                Mem_Free(Crunched_Buffer);

                Hide_Menu(Menu_Saving, Pointer);
                Destroy_Menu(Menu_Saving);
            }
            break;

        case PAD_CANCEL:

            Hide_Menu(Menu_Ask_Saving, Pointer);
            Destroy_Menu(Menu_Ask_Saving);
            break;
    }
}

/* --------------------------------
   Ask the player if the game have to be saved or not
   -------------------------------- */
void Save_Game(int Cur_Location)
{
    memset(&Save, 0, sizeof(SAVEDAT));
    Save.Save_Location = (UBYTE) Cur_Location;

    Menu_Ask_Saving = Construct_Menu(Font,
                                 Pic_Menus,
                                 Ask_Saving_Events,
                                 MENU_TEXT_CENTER,
                                 4, 3,
                                 2, 10,
                                 MOVE_HORIZONTALLY,
                                 "THIS WILL ERASE\nANY PREVIOUSLY SAVED GAME\n \nCONTINUE   STOP\n"
                                );
    Show_Menu(Menu_Ask_Saving, Pointer, -1, 190, 5 * 8);
}

/* --------------------------------
   Restore a saved game
   -------------------------------- */
BOOL Restore_Game()
{
    int i;
    LPSAVEDAT Unpacked_Data;
    int Save_Location;

    Unpacked_Data = (LPSAVEDAT) Depack_Data(Restored_Game, sizeof(SAVEDAT));
    if(Unpacked_Data)
    {
        memcpy(&Party, &Unpacked_Data->Party, sizeof(PARTY));
        /* Restore the current stock of all towns */
        for(i = 0; i < MAX_TOWNS; i++)
        {
            memcpy(&Towns[i].Stock, &Unpacked_Data->Towns_Stock[i][0], MAX_ITEMS);
        }
        Save_Location = Unpacked_Data->Save_Location;
        Mem_Free(Unpacked_Data);

        if(Save_Location & 0x80)
        {
            /* TODO: Was saved from a dungeon, restore in wilderness
                     so go to the wilderness part setting a position nearby the dungeon
            */
            return TRUE;
        }
        else
        {
            /* Was saved from a town */
            Current_Sequence = SEQUENCE_TOWN;
            Set_Fade(Current_Palette, 16, MOVE_NONE);
            if(!Init_Town(Save_Location)) return TRUE;
            return TRUE;
        }
            
    }
    return FALSE;
}

/* --------------------------------
   Check if a game is present in the NVRAM
   -------------------------------- */
void Check_Save_Game()
{
    /* Warn user about any NVRAM issues if necessary */
    switch(Got_NVRAM)
    {
        case NVRAM_FAIL:

            Menu_NVRAM_Warning = Construct_Menu(Font,
                                                Pic_Menus,
                                                NULL,
                                                MENU_TEXT_CENTER,
                                                4, 0,
                                                1, 0,
                                                MOVE_NONE,
                                                "%s\n \n%s\n%s",
                                                "CAUTION !",
                                                "NO NVRAM DETECTED",
                                                "GAME SAVING DISABLED"
                                               );
            Can_Save = FALSE;
            break;

        case NVRAM_OK:

            /* Preload the packed data */
            if(NVRAM_Load("P", "2", Restored_Game, 1024))
            {
                /* Got our saved game data */
                Saved_Game = TRUE;
            }
            break;
    }

    /* If we don't find one of our save game inside the NVRAM
       we warn the user of the big amount of bytes taken
       by something else than us */
    if(Got_NVRAM != NVRAM_OK &&
       Saved_Game == FALSE &&
       Menu_NVRAM_Warning == NULL)
    {
        Menu_NVRAM_Warning = Construct_Menu(Font,
                                            Pic_Menus,
                                            NULL,
                                            MENU_TEXT_CENTER,
                                            4, 0,
                                            1, 0,
                                            MOVE_NONE,
                                            "%s\n \nONLY %d BYTES LEFT IN NVRAM\n%s",
                                            "CAUTION !",
                                            Got_NVRAM,
                                            "GAME SAVING DISABLED"
                                           );
        Can_Save = FALSE;
    }

    if(Menu_NVRAM_Warning)
    {
        Set_Palette(Current_Palette, 0, SCREEN_COLORS);
        Show_Menu(Menu_NVRAM_Warning, NULL, -1, -1, 0);
        Wait_Pad_Ok_Cancel();
        Hide_Menu(Menu_NVRAM_Warning, NULL);
        Destroy_Menu(Menu_NVRAM_Warning);
    }
}
