/* --------------------------------
   Phantasie II CD32
   Gfx support code
   --------------------------------
   Copyright (C) 2010-2016 Franck Charlet.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:

       1. Redistributions of source code must retain the above copyright notice,
          this list of conditions and the following disclaimer.
       2. Redistributions in binary form must reproduce the above copyright notice,
          this list of conditions and the following disclaimer in the documentation
          and/or other materials provided with the distribution.

   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS "AS IS" AND
   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
   ARE DISCLAIMED. IN NO EVENT SHALL FRANCK CHARLET OR CONTRIBUTORS BE LIABLE
   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
   SUCH DAMAGE.
   -------------------------------- */

/* --------------------------------
   Includes
   -------------------------------- */
#include "startup.h"

/* --------------------------------
   Variables
   -------------------------------- */
extern const WORD SCREEN_WIDTH;
extern const WORD SCREEN_HEIGHT;
extern const WORD SCREEN_BPP;
struct ColorMap *ColorMap;
struct ViewPort ViewPort;
struct RastPort RastPort;
struct cprlist *CopperList_Front;
struct RasInfo RasInfo;
extern LPBITMAP BitMap_Ptr_Front;
extern LPBITMAP BitMap_Ptr_Back;
LPBITMAP BitMap_Front;
LPBITMAP BitMap_Back;
PALETTE32 Main_Palette;
//FLOAT Palette_Coeffs_R[256];
//FLOAT Palette_Coeffs_G[256];
//FLOAT Palette_Coeffs_B[256];
int _Palette_R[256];
int _Palette_G[256];
int _Palette_B[256];
BOOL Got_Blitter;
extern BOOL Even_Frame;

/* --------------------------------
   Private functions
   -------------------------------- */
extern "C" void RLEDepack(int Size, void *Source, void *Dest);
void Load_Palette(WORD Start, WORD Amount);
void Clear_Array(WORD X, WORD Y, WORD Width, WORD Height);

/* --------------------------------
   Create an array for graphical operations
   -------------------------------- */
LPBITMAP Create_Bitmap(WORD Width,
                       WORD Height,
                       WORD Depth,
                       BOOL Clear,
                       struct BitMap *Mask)
{
    int i;
    LPBITMAP Bitm;
    
    WaitBlit();
    Bitm = (LPBITMAP) Mem_Alloc(sizeof(BITMAP));
    if(Bitm == NULL) return NULL;

    InitBitMap(&Bitm->Bmp, Depth, Width, Height);
    Bitm->Width = Width;
    Bitm->Height = Height;
    Bitm->Depth = Depth;
    Bitm->Mask = Mask;

    /* Align */
    Width = RowBytes(Width) << 3;
    for(i = 0; i < Depth; i++)
    {
        Bitm->Bmp.Planes[i] = AllocRaster(Width, Height);
        if(!Bitm->Bmp.Planes[i])
        {
            return NULL;
        }
        if(Clear) BltClear(Bitm->Bmp.Planes[i], (Width >> 3) * Height, 1);
    }
    return Bitm;
}

/* --------------------------------
   Free a graphical array
   -------------------------------- */
void _Destroy_Bitmap(LPBITMAP Bitm)
{
    int i;

    if(Bitm)
    {
        /* Automatically frees the mask plane if there's one */
        if(Bitm->Mask) FreeBitMap(Bitm->Mask);

        for(i = 0; i < Bitm->Depth; i++)
        {
            if(Bitm->Bmp.Planes[i]) FreeRaster(Bitm->Bmp.Planes[i], Bitm->Width, Bitm->Height);
        }
        Mem_Free(Bitm);
    }
}

/* --------------------------------
   Return the current BitMap structure of the screen
   -------------------------------- */
struct BitMap *Get_Screen_BitMap()
{
    return RastPort.BitMap;
}

/* --------------------------------
   Create a mask for a given bitmap
   -------------------------------- */
int Create_Bitmap_Mask(LPBITMAP BitMap)
{
    struct BitMap *Mask;
    UBYTE *ptrMask;
    UBYTE *ptrBp;
    int i;
    int j;
    int Bytes;

    /* This should not happen under normal circumstances */
    if(BitMap->Mask)
    {
        FreeBitMap(BitMap->Mask);
    }
    BitMap->Mask = NULL;

    /* Allocate the plane for the Mask */
    Mask = AllocBitMap(BitMap->Width, BitMap->Height, 1, BMF_CLEAR, NULL);
    Bytes = Mask->BytesPerRow;
    WaitBlit();
    if(Mask)
    {
        BitMap->Mask = Mask;
        for(i = 0; i < BitMap->Depth; i++)
        {
            ptrBp = BitMap->Bmp.Planes[i];
            ptrMask = BitMap->Mask->Planes[0];
            if(i == 0)
            {
                for(j = 0; j < Bytes * BitMap->Height; j++)
                {
                    *ptrMask++ = *ptrBp++;
                }
            }
            else
            {
                for(j = 0; j < Bytes * BitMap->Height; j++)
                {
                    *ptrMask++ |= *ptrBp++;
                }
            }
        }
        return TRUE;
    }
    return FALSE;
}

/* --------------------------------
   Clear an array
   -------------------------------- */
void Clear_Array(WORD X, WORD Y, WORD Width, WORD Height)
{
    EraseRect(&RastPort, X, Y, X + Width - 1, Y + Height - 1);
}

/* --------------------------------
   Fill an array
   -------------------------------- */
void Fill_Array(WORD X, WORD Y, WORD Width, WORD Height)
{
    RectFill(&RastPort, X, Y, X + Width - 1, Y + Height - 1);
}

/* --------------------------------
   Copy a bitmap onto the screen
   -------------------------------- */
void Copy_To_Screen(LPBITMAP Src,
                    WORD Src_X,
                    WORD Src_Y,
                    WORD Dst_X,
                    WORD Dst_Y,
                    WORD Width,
                    WORD Height)
{
    int Real_Width = Width;
    int Real_Height = Height;
    
    if(Dst_X < 0)
    {
        Real_Width += Dst_X;
        Src_X -= Dst_X;
        Dst_X = 0;
    }
    else
    {
        if(((SCREEN_WIDTH - Dst_X)) < Real_Width) Real_Width = (SCREEN_WIDTH - Dst_X);
    }
    if(Dst_Y < 0)
    {
        Real_Height += Dst_Y;
        Src_Y -= Dst_Y;
        Dst_Y = 0;
    }
    else        
    {
        if(((SCREEN_HEIGHT - Dst_Y) ) < Real_Height) Real_Height = (SCREEN_HEIGHT - Dst_Y);
    }

    if(Real_Width > 0 && Real_Height > 0)
    {
        if(Src->Mask)
        {
            WaitBlit();
            BltMaskBitMapRastPort(&Src->Bmp,
                                  Src_X,
                                  Src_Y,
                                  &RastPort,
                                  Dst_X,
                                  Dst_Y,
                                  Real_Width,
                                  Real_Height,
                                  ABC | ABNC | ANBC,
                                  Src->Mask->Planes[0]);
        }
        else
        {
            WaitBlit();
            BltBitMapRastPort(&Src->Bmp,
                              Src_X,
                              Src_Y,
                              &RastPort,
                              Dst_X,
                              Dst_Y,
                              Real_Width,
                              Real_Height,
                              ABC | ABNC);
        }
    }
}

/* --------------------------------
   Clear both screen buffers
   -------------------------------- */
void Clear_Screen(WORD X,
                  WORD Y,
                  WORD Width,
                  WORD Height)
{
    Clear_Array(X, Y, Width, Height);
    Swap_Physical_Buffers();
    Clear_Array(X, Y, Width, Height);
}

/* --------------------------------
   Copy a static bitmap onto both screen pages
   -------------------------------- */
void Static_To_Screen(LPBITMAP Src,
                      WORD Src_X,
                      WORD Src_Y,
                      WORD Dst_X,
                      WORD Dst_Y,
                      WORD Width,
                      WORD Height)
{
    Copy_To_Screen(Src, Src_X, Src_Y, Dst_X, Dst_Y, Width, Height);
    Swap_Physical_Buffers();
    Copy_To_Screen(Src, Src_X, Src_Y, Dst_X, Dst_Y, Width, Height);
}

/* --------------------------------
   Copy a bitmap onto another bitmap
   -------------------------------- */
void Copy_To_Bitmap(LPBITMAP Src,
                    WORD Src_X,
                    WORD Src_Y,
                    LPBITMAP Dst,
                    WORD Dst_X,
                    WORD Dst_Y,
                    WORD Width,
                    WORD Height,
                    UBYTE Minterms)
{
    int Real_Width = Width;
    int Real_Height = Height;
    int Real_Width2 = Width;
    int Real_Height2 = Height;
    
    if(Src_X < 0) Src_X = 0;
    if(Src_Y < 0) Src_Y = 0;
    if(Dst_X < 0) Dst_X = 0;
    if(Dst_Y < 0) Dst_Y = 0;
    
    if((Src->Width - Src_X) < Real_Width) Real_Width = (Src->Width - Src_X);
    if((Src->Height - Src_Y) < Real_Height) Real_Height = (Src->Height - Src_Y);

    if((Dst->Width - Dst_X) < Real_Width2) Real_Width2 = (Dst->Width - Dst_X);
    if((Dst->Height - Dst_Y) < Real_Height2) Real_Height2 = (Dst->Height - Dst_Y);

    /* Take the smallest one */
    if(Real_Width > Real_Width2) Real_Width = Real_Width2;
    if(Real_Height > Real_Height2) Real_Height = Real_Height2;

    if(Real_Width > 0 && Real_Height > 0)
    {
        WaitBlit();
        BltBitMap(&Src->Bmp,
                  Src_X,
                  Src_Y,
                  &Dst->Bmp,
                  Dst_X,
                  Dst_Y,
                  Real_Width,
                  Real_Height,
                  Minterms,
                  0xff,
                  NULL);
    }
}

/* --------------------------------
   Scroll a part of the screen
   -------------------------------- */
void Scroll(WORD X,
            WORD Y,
            WORD Width,
            WORD Height,
            WORD Dx,
            WORD Dy)
{
    WaitBlit();
    ScrollRaster(&RastPort, Dx, Dy, X, Y, X + Width - 1, Y + Height - 1);
}

/* --------------------------------
   Move the whole screen
   -------------------------------- */
void Move_Screen(WORD X, WORD Y)
{
    ViewPort.DxOffset = X;
    ViewPort.DyOffset = Y;
    ScrollVPort(&ViewPort);
}

/* --------------------------------
   Draw a line
   -------------------------------- */
void Draw_Line(WORD X1,
               WORD Y1,
               WORD X2,
               WORD Y2,
               UBYTE Color)
{
    WaitBlit();
    SetAPen(&RastPort, Color);
    Move(&RastPort, X1, Y1);
    Draw(&RastPort, X2, Y2);
}

/* --------------------------------
   Install a new palette into the viewport
   -------------------------------- */
void Set_Palette(ULONG *Palette, WORD Start, WORD Amount)
{
    int i;
    ULONG R;
    ULONG G;
    ULONG B;
    LPRGB32 Palette_Dst_Colors;

    Palette_Dst_Colors = &Main_Palette.Palette[Start];

    for(i = 0; i < Amount; i++)
    {
        R = ((Palette[i] & 0xff0000) << 8);
        G = ((Palette[i] & 0x00ff00) << 16);
        B = ((Palette[i] & 0x0000ff) << 24);
        Palette_Dst_Colors[i].R = R;
        Palette_Dst_Colors[i].G = G;
        Palette_Dst_Colors[i].B = B;
    }
    Load_Palette(Start, Amount);
}

/* --------------------------------
   Load the main palette into the viewport
   -------------------------------- */
void Load_Palette(WORD Start, WORD Amount)
{
    Main_Palette.Amount = (Amount << 16) + Start;
    /* We need an empty entry */
    Main_Palette.Palette[Amount].R = 0;
    Main_Palette.Palette[Amount].G = 0;
    Main_Palette.Palette[Amount].B = 0;
    LoadRGB32(&ViewPort, (ULONG *) &Main_Palette);
}

/* --------------------------------
   Pre-calculate the coefficient that will be needed to perform the fading
   -------------------------------- */
void Calc_Fade_Palette(ULONG *Palette,
                       WORD Start,
                       WORD Amount
                       /*,int Speed*/)
{
    int i;
    LONG Src_Color;
    int Src_R;
    int Src_G;
    int Src_B;
/*    int Dst_R;
    int Dst_G;
    int Dst_B;
    int Delta_R;
    int Delta_G;
    int Delta_B;
    FLOAT Scaler;
    FLOAT FltSpeed;
    LPRGB32 Local_Palette = &Main_Palette.Palette[Start];*/

//    Scaler = IEEESPFlt(0xff);
  //  FltSpeed = IEEESPFlt(Speed);

    for(i = 0; i < Amount; i++)
    {
        Src_Color = Palette[i];
        Src_R = (Src_Color & 0xff0000) >> 16;
        Src_G = (Src_Color & 0xff00) >> 8;
        Src_B = Src_Color & 0xff;
        _Palette_R[i] = Src_R;
        _Palette_G[i] = Src_G;
        _Palette_B[i] = Src_B;

/*        Dst_R = (Local_Palette[i].R >> 24);
        Dst_G = (Local_Palette[i].G >> 24);
        Dst_B = (Local_Palette[i].B >> 24);

        Delta_R = (Src_R - Dst_R);
        Delta_G = (Src_G - Dst_G);
        Delta_B = (Src_B - Dst_B);

        Palette_Coeffs_R[i] = IEEESPMul(FltSpeed, IEEESPDiv(IEEESPFlt(Delta_R), Scaler));
        Palette_Coeffs_G[i] = IEEESPMul(FltSpeed, IEEESPDiv(IEEESPFlt(Delta_G), Scaler));
        Palette_Coeffs_B[i] = IEEESPMul(FltSpeed, IEEESPDiv(IEEESPFlt(Delta_B), Scaler));*/
    }
}

/* --------------------------------
   Fade between a given palette and the screen palette
   -------------------------------- */
BOOL Fade_Main_Palette(ULONG *Palette,
                       WORD Start,
                       WORD Amount,
                       int Speed)
{
    int Src_R;
    int Src_G;
    int Src_B;
    int Dst_R;
    int Dst_G;
    int Dst_B;
    ULONG Dst_Orig_R;
    ULONG Dst_Orig_G;
    ULONG Dst_Orig_B;
    //FLOAT Scale_R;
    //FLOAT Scale_G;
    //FLOAT Scale_B;
    int i;

    BOOL Modified = FALSE;
    LPRGB32 Local_Palette = &Main_Palette.Palette[Start];

    for(i = 0; i < Amount; i++)
    {
        Src_R = _Palette_R[i];
        Dst_Orig_R = Local_Palette[i].R;
        Dst_R = Dst_Orig_R >> 24;
        if(Src_R != Dst_R)
        {
            //Scale_R = Palette_Coeffs_R[i];
            if(Src_R > Dst_R)
            {
                Dst_R += Speed; //= IEEESPFix(IEEESPAdd(IEEESPFlt(Dst_R), Scale_R));
                if(Dst_R > Src_R) Dst_R = Src_R;
            }
            else
            {
                Dst_R -= Speed; //= IEEESPFix(IEEESPAdd(IEEESPFlt(Dst_R), Scale_R));
                if(Dst_R < Src_R) Dst_R = Src_R;
            }
        }

        Src_G = _Palette_G[i];
        Dst_Orig_G = Local_Palette[i].G;
        Dst_G = Dst_Orig_G >> 24;
        if(Src_G != Dst_G)
        {
            //Scale_G = Palette_Coeffs_G[i];
            if(Src_G > Dst_G)
            {
                Dst_G += Speed; //= IEEESPFix(IEEESPAdd(IEEESPFlt(Dst_G), Scale_G));
                if(Dst_G > Src_G) Dst_G = Src_G;
            }
            else
            {
                Dst_G -= Speed; //= IEEESPFix(IEEESPAdd(IEEESPFlt(Dst_G), Scale_G));
                if(Dst_G < Src_G) Dst_G = Src_G;
            }
        }

        Src_B = _Palette_B[i];
        Dst_Orig_B = Local_Palette[i].B;
        Dst_B = Dst_Orig_B >> 24;
        if(Src_B != Dst_B)
        {
            //Scale_B = Palette_Coeffs_B[i];
            if(Src_B > Dst_B)
            {
                Dst_B += Speed; //= IEEESPFix(IEEESPAdd(IEEESPFlt(Dst_B), Scale_B));
                if(Dst_B > Src_B) Dst_B = Src_B;
            }
            else
            {
                Dst_B -= Speed; //= IEEESPFix(IEEESPAdd(IEEESPFlt(Dst_B), Scale_B));
                if(Dst_B < Src_B) Dst_B = Src_B;
            }
        }
        Local_Palette[i].R = Dst_R << 24;
        Local_Palette[i].G = Dst_G << 24;
        Local_Palette[i].B = Dst_B << 24;
        if(Dst_Orig_R != Local_Palette[i].R ||
           Dst_Orig_G != Local_Palette[i].G ||
           Dst_Orig_B != Local_Palette[i].B)
        {
            Modified = TRUE;
        }
    }
    if(Modified)
    {
        Load_Palette(Start, Amount);
    }
    return Modified;
}

/* --------------------------------
   Fade a palette
   -------------------------------- */
void Fade_Palette(ULONG *Palette_Src,
                  ULONG *Palette_Dst,
                  WORD Start,
                  WORD Amount,
                  int Speed)
{
    LONG Color;
    LONG Src_R;
    LONG Src_G;
    LONG Src_B;
    LONG Dst_R;
    LONG Dst_G;
    LONG Dst_B;
    int i;

    for(i = Start; i < Start + Amount; i++)
    {
        Color = Palette_Src[i];
        Src_R = (Color & 0xff0000) >> 16;
        Src_G = (Color & 0xff00) >> 8;
        Src_B = Color & 0xff;
    
        Color = Palette_Dst[i];
        Dst_R = (Color & 0xff0000) >> 16;
        Dst_G = (Color & 0xff00) >> 8;
        Dst_B = Color & 0xff;

        if(Src_R != Dst_R)
        {
            if(Src_R > Dst_R)
            {
                Dst_R += Speed;
                if(Dst_R > Src_R) Dst_R = Src_R;
            }
            else
            {
                Dst_R -= Speed;
                if(Dst_R < Src_R) Dst_R = Src_R;
            }
        }
        if(Src_G != Dst_G)
        {
            if(Src_G > Dst_G)
            {
                Dst_G += Speed;
                if(Dst_G > Src_G) Dst_G = Src_G;
            }
            else
            {
                Dst_G -= Speed;
                if(Dst_G < Src_G) Dst_G = Src_G;
            }
        }
        if(Src_B != Dst_B)
        {
            if(Src_B > Dst_B)
            {
                Dst_B += Speed;
                if(Dst_B > Src_B) Dst_B = Src_B;
            }
            else
            {
                Dst_B -= Speed;
                if(Dst_B < Src_B) Dst_B = Src_B;
            }
        }
        Palette_Dst[i] = (Dst_R << 16) | (Dst_G << 8) | (Dst_B);
    }
}

/* --------------------------------
   Get exclusive rights over the blitter
   -------------------------------- */
void Lock_Blitter()
{
    if(!Got_Blitter)
    {
        OwnBlitter();
        WaitBlit();
        Got_Blitter = TRUE;
    }
}

/* --------------------------------
   Free the blitter for other tasks
   -------------------------------- */
void Unlock_Blitter()
{
    if(Got_Blitter)
    {
        WaitBlit();
        DisownBlitter();
        Got_Blitter = FALSE;
    }
}

/* --------------------------------
   Create a fontset from a bitmap
   -------------------------------- */
LPFONT Create_Font(LPBITMAP Source,
                   UBYTE *Sizes,
                   WORD Src_X,
                   WORD Src_Y,
                   WORD Width,
                   WORD Height)
{
    LPFONT Font;

    Font = (LPFONT) Mem_Alloc(sizeof(FONT));
    if(Font)
    {
        Font->Sizes = Sizes;
        Font->Bank = Source;
        Font->Src_X = Src_X;
        Font->Src_Y = Src_Y;
        Font->Width = Width;
        Font->Height = Height;
    }
    return Font;
}

/* --------------------------------
   Free a fontset object
   -------------------------------- */
void _Destroy_Font(LPFONT Font)
{
    Mem_Free(Font);
}

/* --------------------------------
   Get the length of a text in pixels
   -------------------------------- */
int Get_Text_Length(LPFONT Font,
                    char *Text,
                    int Len,
                    int *Size,
                    int *Chars)
{
    int i;
    //int j;
    char Car;

    /* Display all chars */
    *Size = 0;
    *Chars = 0;
    i = 0;
    while(i < Len)
    {
        Car = Text[i];
        if(Car != 13)
        {
            if(Car == 10)
            {
                /* There's more */
                return i + 1;
            }
            else
            {
                //for(j = 0; j < Font->ASCII_Len; j++)
                //{
                    /*if(Font->ASCII[j] == Car)
                    {
                        if(Font->Sizes)
                        {
                            *Size += Font->Sizes[j];
                        }
                        else
                        {*/
                            /* Fixed size */
                            *Size += Font->Width;
                        //}
                        *Chars += 1;
                //        break;
                 //   }
               // }
            }
        }
        i++;
    }
    /* End of text */
    return 0;
}

/* --------------------------------
   Print a text on the screen
   -------------------------------- */
void Print(LPFONT Font,
           int Flags,
           WORD Pos_X,
           WORD Pos_Y,
           WORD Max_X,
           int Static,
           UBYTE Minterms,
           char *Format,
           ...)
{
    char szBuffer[512];
    int Len;
    int i;
//   int j;
    int Size_X;
    int Size_Chars;
    va_list Arg;
    char Car;
    WORD Orig_Pos_X;
    int Font_Offset;
    int Max_Width = 0;
    int Max_Height = Font->Height + 1;
    int Def_Pos_X;
    int Def_Pos_Y;

    Font_Offset = Font->Src_Y + (Flags & MENU_TEXT_SECONDARY_FONT ? Font->Height : 0);

    Def_Pos_X = Pos_X;
    Def_Pos_Y = Pos_Y;

    va_start(Arg, Format);
        Len = vsprintf(szBuffer, Format, (char *) Arg);

        if(Len > 0)
        {
            Orig_Pos_X = Pos_X;

            i = 0;
            if(Pos_X == -1)
            {
                Get_Text_Length(Font, &szBuffer[i], Len - i, &Size_X, &Size_Chars);
                Pos_X = (Max_X - Size_X) / 2;
                Def_Pos_X = Pos_X;
            }
        
            /* Display all chars */
            while(i < Len)
            {
                Car = szBuffer[i];
                if(Car != 13)
                {
                    if(Car == 10)
                    {
                        i++;
                        /* Next line */
                        Pos_X = Orig_Pos_X;
                        if(Pos_X == -1)
                        {
                            Get_Text_Length(Font, &szBuffer[i], Len - i, &Size_X, &Size_Chars);
                            Pos_X = (Max_X - Size_X) / 2;
                        }
                        i--;
                        Pos_Y += Font->Height + 1;
                        Max_Height += Font->Height + 1;
                    }
                    else
                    {
                        Car -= 0x20;
                        /* Look for the char inside the list
                           and display it if found */
                        //for(j = 0; j < Font->ASCII_Len; j++)
                        //{
                            //if(Font->ASCII[j] == Car)
                            //{
                                Copy_To_Screen(Font->Bank,
                                               Font->Src_X + (Car * Font->Width),
                                               Font_Offset,
                                               Pos_X,
                                               Pos_Y,
                                               Font->Width,
                                               Font->Height);
                                Max_Width += Font->Width;
                                /* Next pos */
                                /*if(Font->Sizes)
                                {
                                    // Proportional size
                                    Pos_X += Font->Sizes[j];
                                }
                                else
                                {*/
                                    // Fixed size
                                    Pos_X += Font->Width;
                                //}
                                //break;
                            //}
                        //}
                    }
                }
                i++;
            }
        }
        if(Static)
        {
            Swap_Physical_Buffers();

            Copy_To_Bitmap(BitMap_Ptr_Back,
                           Def_Pos_X,
                           Def_Pos_Y,
                           BitMap_Ptr_Front,
                           Def_Pos_X,
                           Def_Pos_Y,
                           Max_Width,
                           Max_Height,
                           ABC | ABNC
                          );
        }
    va_end(Arg);
}

/* --------------------------------
   Load an IFF picture and return a bitmap structure from it
   -------------------------------- */
LPBITMAP Load_Picture(char *FileName,
                      BOOL Create_Mask,
                      ULONG *Palette,
                      int Amount_Colors)
{
    struct IFFHandle *Context;
    struct StoredProperty *Property;
    BPTR Handle;
    LPIFFHEADER bmhd;
    LPIFFPALETTE cmap;
    int BodySize;
    int Read_BodySize;
    int Width;
    int Bytes;
    BYTE *ptrBp;
    LPBITMAP BitMap = NULL;
    UBYTE *ptrLinearBp;
    BYTE *LinearPlanes = NULL;
    BYTE *Body = NULL;
    int i;
    int j;
    int k;

    Context = AllocIFF();
    if(Context)
    {
        /* Open the file */
        Handle = F_Open((STRPTR) FileName, MODE_OLDFILE);
        if(Handle)
        {
            Context->iff_Stream = Handle;
            InitIFFasDOS(Context);
            if(!OpenIFF(Context, IFFF_READ))
            {
                PropChunk(Context, ID_ILBM, ID_BMHD);       /* Infos */
                if(Palette)
                {
                    PropChunk(Context, ID_ILBM, ID_CMAP);   /* Palette */
                }
                StopChunk(Context, ID_ILBM, ID_BODY);
                if(!ParseIFF(Context, IFFPARSE_SCAN))
                {
                    /* Obtain header */
                    bmhd = (LPIFFHEADER) FindProp(Context, ID_ILBM, ID_BMHD)->sp_Data;
                    /* Align */
                    Width = RowBytes(bmhd->w) << 3;

                    /* Create the structure to hold it */
                    BitMap = Create_Bitmap(Width, bmhd->h, bmhd->nPlanes, TRUE, NULL);
                    if(BitMap)
                    {
                        Bytes = BitMap->Width >> 3;
                        BodySize = CurrentChunk(Context)->cn_Size;
                        LinearPlanes = (BYTE *) Mem_Alloc(Bytes * BitMap->Height * BitMap->Depth);
                        Body = (BYTE *) Mem_Alloc(BodySize);
                        Read_BodySize = 0;
                        if(Body)
                        {
                            Read_BodySize = ReadChunkBytes(Context, Body, BodySize);
                        }
                        if(LinearPlanes && Body && Read_BodySize == BodySize)
                        {
                            if(bmhd->compression == cmpByteRun1)
                            {
                                /* Depack it into the temporary buffer */
                                RLEDepack(Bytes * BitMap->Height * BitMap->Depth,
                                          Body,
                                          LinearPlanes);
                            }
                            else
                            {
                                /* Straight copy */
                                memcpy(LinearPlanes, Body, BodySize);
                            }
    
                            /* Dispatch the interleaved picture */
                            ptrBp = LinearPlanes;
                            for(k = 0; k < BitMap->Height; k++)
                            {
                                for(i = 0; i < BitMap->Depth; i++)
                                {
                                    ptrLinearBp = BitMap->Bmp.Planes[i] + (k * Bytes);
                                    for(j = 0; j < Bytes; j++)
                                    {
                                        *ptrLinearBp++ = *ptrBp++;
                                    }
                                }
                            }                           

                            /* Create the mask from the picture */
                            if(Create_Mask)
                            {
                                if(Create_Bitmap_Mask(BitMap) == FALSE)
                                {
                                    Mem_Free(Body);
                                    Mem_Free(LinearPlanes);
                                    /* Safety (Mask is freed automatically) */
                                    Destroy_Bitmap(BitMap);
                                    BitMap = NULL;
                                }
                            }
                            if(BitMap)
                            {
                                BitMap->Colors = 0;
                                if(Palette)
                                {
                                    Property = FindProp(Context, ID_ILBM, ID_CMAP);
                                    BitMap->Colors = Property->sp_Size / 3;
                                    cmap = (LPIFFPALETTE) Property->sp_Data;
                                    if(cmap)
                                    {
                                        /* Don't chew more than we can bite */
                                        if(Amount_Colors > BitMap->Colors)
                                        {
                                            Amount_Colors = BitMap->Colors;
                                        }
                                        for(i = 0; i < Amount_Colors; i++)
                                        {
                                            Palette[i] = (cmap->Palette[i].R & 0xff) << 16 |
                                                         (cmap->Palette[i].G & 0xff) << 8 |
                                                         (cmap->Palette[i].B & 0xff);
                                        }
                                    }
                                }
    
                                /* Free our temporary buffers now */
                                Mem_Free(Body);
                                Mem_Free(LinearPlanes);
                            }
                        }
                        else
                        {
                            Mem_Free(Body);
                            Mem_Free(LinearPlanes);
                            /* Safety (Mask is freed automatically) */
                            Destroy_Bitmap(BitMap);
                            BitMap = NULL;
                            Printf("Couldn't open picture: %s\n", FileName);
                        }
                    }
                }
                CloseIFF(Context);
            }
        }
        else
        {
            Printf("Couldn't open picture: %s\n", FileName);
        }
        F_Close(Handle);
        FreeIFF(Context);
    }
    return BitMap;
}

/* --------------------------------
   Create a movable object
   -------------------------------- */
LPOBJECT Create_Object(WORD Width,
                       WORD Height,
                       LPBITMAP Bitmap_Bank)
{
    LPOBJECT Obj;

    Obj = (LPOBJECT) Mem_Alloc(sizeof(OBJECT));
    if(Obj)
    {
        Obj->Width = Width;
        Obj->Height = Height;
        Obj->Saved = FALSE;
        Obj->Old_XA = -1000;
        Obj->Old_YA = -1000;
        Obj->Old_XB = -1000;
        Obj->Old_YB = -1000;
        /* Simply linked to save memory */
        Obj->Bank = Bitmap_Bank;
        /* Back saving bitmap */
        Obj->BackA = Create_Bitmap(Width, Height, SCREEN_BPP, TRUE, NULL);
        Obj->BackB = Create_Bitmap(Width, Height, SCREEN_BPP, TRUE, NULL);
    }
    return Obj;
}

/* --------------------------------
   Free an object resources
   -------------------------------- */
void _Destroy_Object(LPOBJECT Obj)
{
    if(Obj)
    {
        Destroy_Bitmap(Obj->BackA);
        Destroy_Bitmap(Obj->BackB);
        Mem_Free(Obj);
    }
}

/* --------------------------------
   Change the bitmap bank used for an object
   -------------------------------- */
void Set_Object_Bank(LPOBJECT Obj,
                     LPBITMAP Bitmap_Bank)
{
    Obj->Bank = Bitmap_Bank;
}

/* --------------------------------
   Erase an object from the screen
   -------------------------------- */
void Erase_Object(LPOBJECT Obj)
{
    /* Need to wait for both back buffers to be filled */
    if(Obj->Saved >= 2)
    {
        /* Restore backgrounds */
        if(Even_Frame)
        {
            Copy_To_Bitmap(Obj->BackA,
                           0,
                           0,
                           BitMap_Ptr_Front,
                           Obj->Old_XA,
                           Obj->Old_YA,
                           Obj->Width,
                           Obj->Height,
                           ABC | ABNC
                          );
        }
        else
        {
            Copy_To_Bitmap(Obj->BackB,
                           0,
                           0,
                           BitMap_Ptr_Front,
                           Obj->Old_XB,
                           Obj->Old_YB,
                           Obj->Width,
                           Obj->Height,
                           ABC | ABNC
                          );
        }
        Obj->Saved = 2;
    }
}

/* --------------------------------
   Erase a list of objects from the screen
   -------------------------------- */
void Erase_Objects(LPOBJECT Objs, short Amount)
{
    short i;

    for(i = 0 ; i < Amount; i++)
    {
        /* Restore backgrounds */
        if(Even_Frame)
        {
            if(Objs[i].Saved >= 2)
            {
                Copy_To_Bitmap(Objs[i].BackA,
                               0,
                               0,
                               BitMap_Ptr_Front,
                               Objs[i].Old_XA,
                               Objs[i].Old_YA,
                               Objs[i].Width,
                               Objs[i].Height,
                               ABC | ABNC
                              );
                Objs[i].Saved = 2;
            }
        }
        else
        {
            if(Objs[i].Saved >= 2)
            {
                Copy_To_Bitmap(Objs[i].BackB,
                               0,
                               0,
                               BitMap_Ptr_Front,
                               Objs[i].Old_XB,
                               Objs[i].Old_YB,
                               Objs[i].Width,
                               Objs[i].Height,
                               ABC | ABNC
                              );
                Objs[i].Saved = 2;
            }
        }
    }
}

/* --------------------------------
   Display an object on the screen
   -------------------------------- */
void Display_Object(LPOBJECT Obj,
                    WORD Pos_X,
                    WORD Pos_Y,
                    WORD Src_X,
                    WORD Src_Y)
{
    /* Save backgrounds */
    if(Even_Frame)
    {
        Copy_To_Bitmap(BitMap_Ptr_Front,
                       Pos_X,
                       Pos_Y,
                       Obj->BackA,
                       0,
                       0,
                       Obj->Width,
                       Obj->Height,
                       ABC | ABNC
                      );
        Obj->Old_XA = Pos_X;
        Obj->Old_YA = Pos_Y;
    }
    else
    {
        Copy_To_Bitmap(BitMap_Ptr_Front,
                       Pos_X,
                       Pos_Y,
                       Obj->BackB,
                       0,
                       0,
                       Obj->Width,
                       Obj->Height,
                       ABC | ABNC
                      );
        Obj->Old_XB = Pos_X;
        Obj->Old_YB = Pos_Y;
    }
    /* Display the object */
    Copy_To_Screen(Obj->Bank,
                   Src_X, Src_Y,
                   Pos_X, Pos_Y,
                   Obj->Width, Obj->Height);
    Obj->Saved++;
}

/* --------------------------------
   Copy the Front buffer into the back one
   -------------------------------- */
void Copy_From_Back(WORD X,
                    WORD Y,
                    WORD Width,
                    WORD Height)
{
    Copy_To_Bitmap(BitMap_Ptr_Back,
                   X,
                   Y,
                   BitMap_Ptr_Front,
                   X,
                   Y,
                   Width,
                   Height,
                   ABC | ABNC
                  );
}
