/* --------------------------------
   Phantasie II CD32
   Non-volatile RAM support code
   --------------------------------
   Copyright (C) 2010-2016 Franck Charlet.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:

       1. Redistributions of source code must retain the above copyright notice,
          this list of conditions and the following disclaimer.
       2. Redistributions in binary form must reproduce the above copyright notice,
          this list of conditions and the following disclaimer in the documentation
          and/or other materials provided with the distribution.

   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS "AS IS" AND
   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
   ARE DISCLAIMED. IN NO EVENT SHALL FRANCK CHARLET OR CONTRIBUTORS BE LIABLE
   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
   SUCH DAMAGE.
   -------------------------------- */

#ifdef CD32

/* --------------------------------
   Includes
   -------------------------------- */
#include "startup.h"

/* --------------------------------
   Variables
   -------------------------------- */
struct NVBase *NVBase;

/* --------------------------------
   Open nonvolatile.library
   -------------------------------- */
int NVRAM_Init(int Blocks_Required)
{
    struct NVInfo *Infos;
    int Size;

    NVBase = (struct NVBase *) OpenLibrary("nonvolatile.library", 0);
    if(!NVBase) return FALSE;

    Infos = GetNVInfo(TRUE);
    if(Infos)
    {
        /* 1k of RAM (100 blocks) is the bare mininum */
        if(Infos->nvi_MaxStorage < 100)
        {
            FreeNVData(Infos);
            return NVRAM_FAIL;
        }
        
        /* We need to warn the user about anything fishy */
        if((int) Infos->nvi_FreeStorage < Blocks_Required)
        {
            Size = Infos->nvi_FreeStorage * 10;
            FreeNVData(Infos);
            return Size;
        }
    }
    FreeNVData(Infos);
    return NVRAM_OK;
}

/* --------------------------------
   Close nonvolatile.library
   -------------------------------- */
void NVRAM_Close()
{
    if(NVBase) CloseLibrary((struct Library *) NVBase);
}

/* --------------------------------
   Save data into a slot
   -------------------------------- */
int NVRAM_Save(char *Name,
               char *Item,
               UBYTE *Data,
               int Bytes)
{
    if(NVBase)
    {
        return StoreNV(Name,
                       Item,
                       Data,
                       (Bytes / 10) + ((Bytes % 10) ? 1 : 0),
                       TRUE
                      );
    }
    return 0;
}

/* --------------------------------
   Load data from a slot
   -------------------------------- */
BOOL NVRAM_Load(char *Name,
                char *Item,
                UBYTE *Data, 
                int Bytes)
{
    APTR NVData;

    if(NVBase)
    {
        NVData = GetCopyNV(Name,
                           Item,
                           TRUE
                          );
        if(NVData)
        {
            memcpy(Data, NVData, Bytes);
            FreeNVData(NVData);
            return TRUE;
        }
    }
    return FALSE;
}

#endif
